@@ -1,152 +1,12 @@
 import sys
 sys.path.append('.')
-import time, serial, struct, math
+import time, serial, math
+
+from pymycobot.error import MyCobotDataException, check_id
+from pymycobot.common import Command, MyCobotData
 
 
-class Command():
-    #BASIC
-    HEADER = 0xfe
-    FOOTER = 0xfa
-
-    # Overall status
-    POWER_ON = 0x10
-    POWER_OFF = 0x11
-    IS_POWER_ON = 0x12
-    SET_FREE_MODE = 0x13
-
-    # MDI MODE AND OPERATION
-    GET_ANGLES = 0x20
-    SEND_ANGLE = 0x21
-    SEND_ANGLES = 0x22
-    GET_COORDS = 0x23
-    SEND_COORD = 0x24
-    SEND_COORDS = 0x25
-    PAUSE = 0x26
-    IS_PAUSED = 0x27
-    RESUME = 0x28
-    STOP = 0x29
-    IS_IN_POSITION = 0x2a
-    IS_MOVING = 0x2b
-
-    # JOG MODE AND OPERATION
-    JOG_ANGLE = 0x30
-    JOG_COORD = 0x32
-    JOG_STOP = 0x34
-
-    # RUNNING STATUS AND SETTINGS
-    GET_SPEED = 0x40
-    SET_SPEED = 0x41
-    GET_JOINT_MIN_ANGLE = 0x4a
-    GET_JOINT_MAX_ANGLE = 0x4b
-
-    # SERVO CONTROL
-    IS_SERVO_ENABLE = 0x50
-    IS_ALL_SERVO_ENABLE = 0x51
-
-    # ATOM IO
-    SET_COLOR = 0x6a
-    SET_CLAW = 0x66
-
-
-class DataProcesser():
-    # Functional approach
-    def _encode_int8(self, data):
-        return struct.pack('b', data)
-
-    def _encode_int16(self, data):
-        return list(struct.pack('>h', data))
-
-    def _decode_int8(self, data):
-        return struct.unpack('b', data)[0]
-
-    def _decode_int16(self, data):
-        return struct.unpack('>h', data)[0]
-
-    def _angle_to_int(self, angle):
-        return int(angle * 100)
-
-    def _coord_to_int(self, coord):
-        return int(coord * 10)
-
-    def _int_to_angle(self, _int):
-        return round(_int / 100.0, 3)
-
-    def _int_to_coord(self, _int):
-        return round(_int / 10.0, 2)
-
-    def _flatten(self, _list):
-        return sum(([x] if not isinstance(x, list) else self._flatten(x)
-                    for x in _list), [])
-
-    def _process_data(self, data, genre):
-        if not data:
-            return []
-
-        if genre in [
-                Command.SEND_ANGLES, Command.SEND_COORDS,
-                Command.IS_IN_POSITION
-        ]:
-            _data_list = []
-            for value in data[:6]:
-                _data_list.extend(self._encode_int16((value)))
-            for value in data[6:]:
-                _data_list.append((value))
-            return _data_list
-
-        elif genre in [Command.SEND_ANGLE, Command.SEND_COORD]:
-            return self._flatten([
-                data[0],
-                self._flatten(self._encode_int16(data[1])),
-                data[2]
-            ])
-
-        elif genre in [Command.JOG_ANGLE, Command.JOG_COORD]:
-            return self._flatten([
-                data[0], data[1],
-                data[2]
-            ])
-
-        else:
-            return [data[0]]
-
-    def _is_frame_header(self, data, pos):
-        return data[pos] == Command.HEADER and data[pos + 1] == Command.HEADER
-
-    def _process_recived(self, data, genre):
-        if not data:
-            return []
-
-        data = bytearray(data)
-        data_len = len(data)
-        for idx in range(data_len - 1):
-            if self._is_frame_header(data, idx):
-                break
-        else:
-            return []
-
-        data_len = data[idx + 2] - 2
-        cmd_id = data[idx + 3]
-        if cmd_id != genre:
-            return []
-        data_pos = idx + 4
-        avild_data = (data[data_pos:data_pos + data_len])
-
-        res = []
-        if data_len == 12:
-            for idx in range(0, len(avild_data), 2):
-                a = avild_data[idx:idx + 2]
-                res.append(self._decode_int16(a))
-        elif data_len == 2:
-            res.append(self._decode_int16(avild_data))
-        else:
-            res.append(self._decode_int8(avild_data))
-        return res
-
-    def _process_bool(self, data):
-        return data[0] if data else -1
-
-
-class MyCobot(DataProcesser):
+class MyCobot(MyCobotData):
     '''MyCobot Python API
 
     Possessed function:
@@ -186,10 +46,22 @@ class MyCobot(DataProcesser):
         # Servo control
             is_servo_enable()
             is_all_servo_enable()
+            release_servo()
+            focus_servo()
 
         # Atom IO
             set_led_color()
             set_claw()
+            set_pin_mode()
+            set_digital_output()
+            get_digital_input()
+            set_pwm_mode()
+            set_pwm_output()
+            get_gripper_value()
+            set_gripper_state()
+            set_gripper_value()
+            set_gripper_ini()
+            is_gripper_moving()
     '''
 
     def __init__(self, port, boudrate='115200', timeout=0.1):
@@ -221,7 +93,7 @@ class MyCobot(DataProcesser):
         return data
 
     def __mesg(self, genre, data=None, has_reply=False):
-        command_data = self._process_data(data, genre)
+        command_data = self._process_data_command(data, genre)
         LEN = len(command_data) + 2
         command = [
             Command.HEADER, Command.HEADER, LEN, genre, command_data,
@@ -250,8 +122,8 @@ class MyCobot(DataProcesser):
             -1: error data
 
         '''
-        received = self.__mesg(Command.IS_POWER_ON, has_reply=True)
-        return self._process_bool(received)
+        return self._process_single(
+            self.__mesg(Command.IS_POWER_ON, has_reply=True))
 
     def set_free_mode(self):
         # self._write('fefe0213fa')
@@ -268,6 +140,7 @@ class MyCobot(DataProcesser):
         angles = self.__mesg(Command.GET_ANGLES, has_reply=True)
         return [self._int_to_angle(angle) for angle in angles]
 
+    @check_id
     def send_angle(self, id, degree, speed):
         '''Send one angle
 
@@ -337,6 +210,7 @@ class MyCobot(DataProcesser):
             res.append(self._int_to_angle(received[idx]))
         return res
 
+    @check_id
     def send_coord(self, id, coord, speed):
         '''Send one coord
 
@@ -360,7 +234,7 @@ class MyCobot(DataProcesser):
 
         '''
         if len(coords) != 6:
-            raise Exception('The leght of coords should be 6.')
+            raise MyCobotDataException('The leght of coords should be 6.')
 
         coord_list = []
         for idx in range(3):
@@ -374,7 +248,7 @@ class MyCobot(DataProcesser):
         self.__mesg(Command.PAUSE)
 
     def is_paused(self):
-        return self._process_bool(self.__mesg(Command.IS_PAUSED,
+        return self._process_single(self.__mesg(Command.IS_PAUSED,
                                              has_reply=True))
 
     def resume(self):
@@ -395,7 +269,7 @@ class MyCobot(DataProcesser):
             -1: error data
         '''
         if len(data) != 6:
-            raise Exception('The lenght of coords is not right')
+            raise MyCobotDataException('The lenght of coords is not right')
 
         if id == 1:
             data_list = []
@@ -406,12 +280,12 @@ class MyCobot(DataProcesser):
         elif id == 0:
             data_list = [self._coord_to_int(i) for i in data]
         else:
-            raise Exception("id is not right, please input 0 or 1")
+            raise MyCobotDataException("id is not right, please input 0 or 1")
 
         received = self.__mesg(Command.IS_IN_POSITION,
                               data=data_list,
                               has_reply=True)
-        return self._process_bool(received)
+        return self._process_single(received)
 
     def is_moving(self):
         '''
@@ -421,22 +295,26 @@ class MyCobot(DataProcesser):
             1 : is moving
             -1: error data
         '''
-        return self._process_bool(self.__mesg(Command.IS_MOVING,
+        return self._process_single(self.__mesg(Command.IS_MOVING,
                                              has_reply=True))
 
     # JOG mode and operation
+    @check_id
     def jog_angle(self, joint_id, direction, speed):
         '''Joint control
 
+        Args:
             joint_id: string
             direction: int [0, 1]
             speed: int (0 - 100)
         '''
         self.__mesg(Command.JOG_ANGLE, data=[joint_id, direction, speed])
 
+    @check_id
     def jog_coord(self, coord_id, direction, speed):
         '''Coord control 
 
+        Args:
             coord: string
             direction: int [0, 1]
             speed: int (0 - 100)
@@ -448,7 +326,7 @@ class MyCobot(DataProcesser):
 
     # Running status and Settings
     def get_speed(self):
-        return self._process_bool(self.__mesg(Command.GET_SPEED,
+        return self._process_single(self.__mesg(Command.GET_SPEED,
                                              has_reply=True))
 
     def set_speed(self, speed):
@@ -458,16 +336,19 @@ class MyCobot(DataProcesser):
             speed (int): 0 - 100
         '''
         if not 0 <= speed <= 100:
-            raise Exception('speed value not right, should be 0 ~ 100 ')
+            raise MyCobotDataException(
+                'speed value not right, should be 0 ~ 100 ')
 
         self.__mesg(Command.SET_SPEED, data=[speed])
 
+    @check_id
     def get_joint_min_angle(self, joint_id):
         angle = self.__mesg(Command.GET_JOINT_MIN_ANGLE,
                            data=[joint_id],
                            has_reply=True)
         return self._int_to_angle(angle[0]) if angle else 0
 
+    @check_id
     def get_joint_max_angle(self, joint_id):
         angle = self.__mesg(Command.GET_JOINT_MAX_ANGLE,
                            data=[joint_id],
@@ -475,14 +356,33 @@ class MyCobot(DataProcesser):
         return self._int_to_angle(angle[0]) if angle else 0
 
     # Servo control
+    @check_id
     def is_servo_enable(self, servo_id):
-        return self._process_bool(
+        return self._process_single(
             self.__mesg(Command.IS_SERVO_ENABLE, data=[servo_id]))
 
     def is_all_servo_enable(self):
-        return self._process_bool(
+        return self._process_single(
             self.__mesg(Command.IS_ALL_SERVO_ENABLE, has_reply=True))
 
+    def release_servo(self, servo_id):
+        '''Power on designated servo
+
+        Args:
+            servo_id: 1 ~ 6
+
+        '''
+        self.__mesg(Command.RELEASE_SERVO, servo_id)
+
+    def focus_servo(self, servo_id):
+        '''Power off designated servo
+
+        Args:
+            servo_id: 1 ~ 6
+
+        '''
+        self.__mesg(Command.RELEASE_SERVO, servo_id)
+
     # Atom IO
     def set_led_color(self, rgb):
         '''Set the light color
@@ -491,7 +391,11 @@ class MyCobot(DataProcesser):
             rgs (str): example 'ff0000'
 
         '''
-        # rgb = struct.pack('bbb', rgb)
+        if len(rgb) != 6:
+            raise MyCobotDataException('rgb format should be like: "FF0000"')
+
+        # data = list(bytearray(rgb))
+        # self.__mesg(Command.SET_COLOR, data=data)
         command = 'fefe056a{}fa'.format(rgb)
         if self._version == 2:
             command = command.decode('hex')
@@ -501,14 +405,81 @@ class MyCobot(DataProcesser):
         self._serial_port.flush()
         time.sleep(0.05)
 
-    def set_claw(self, flag):
-        '''Set claw switch
+    def set_pin_mode(self, pin_no, pin_mode):
+        '''
 
         Args:
-            flag (int): 0 - open, 1 - close
+            pin_no   (int):
+            pin_mode (int): 0 - input, 1 - output, 2 - input_pullup
+
+        '''
+        self.__mesg(Command.SET_PIN_MODE, data=[pin_no, pin_mode])
+
+    def set_digital_output(self, pin_no, pin_signal):
+        '''
+
+        Args:
+            pin_no     (int):
+            pin_signal (int): 0 / 1
+
+        '''
+        self.__mesg(Command.SET_DIGITAL_OUTPUT, data=[pin_no, pin_signal])
+
+    def get_digital_input(self, pin_no):
+        return self._process_single(
+            self.__mesg(Command.GET_DIGITAL_INPUT,
+                        data=[pin_no],
+                        has_reply=True))
+
+    def set_pwm_mode(self, pin_no, channel):
+        self.__mesg(Command.SET_PWM_MODE, data=[pin_no, channel])
+
+    def set_pwm_output(self, channel, pin_val):
+        self.__mesg(Command.SET_PWM_OUTPUT, data=[channel, pin_val])
+
+    def get_gripper_value(self):
+        return self._process_single(
+            self.__mesg(Command.GET_GRIPPER_VALUE, has_reply=True))
+
+    def set_gripper_state(self, flag, speed):
+        '''Set gripper switch
+
+        Args:
+            flag  (int): 0 - open, 1 - close
+            speed (int): 0 ~ 100
 
         '''
         if not flag in [0, 1]:
-            raise Exception('eror flag, please input 0 or 1')
+            raise MyCobotDataException('eror flag, please input 0 or 1')
 
-        self.__mesg(Command.SET_CLAW, data=[flag])
+        self.__mesg(Command.SET_GRIPPER_STATE, data=[flag, speed])
+
+    def set_gripper_value(self, value, speed):
+        '''Set gripper value
+
+        Args:
+            value (int): 0 ~ 496
+            speed (int): 0 ~ 100
+
+        '''
+        self.__mesg(Command.SET_GRIPPER_VALUE, data=[value, speed])
+
+    def set_gripper_ini(self):
+        '''Set the current position to zero
+
+        Current position value is `248`.
+
+        '''
+        self.__mesg(Command.SET_GRIPPER_INI)
+
+    def is_gripper_moving(self):
+        '''Judge whether the gripper is moving or not
+
+        Returns:
+            0 : not moving
+            1 : is moving
+            -1: error data
+
+        '''
+        return self._process_single(
+            self.__mesg(Command.IS_GRIPPER_MOVING, has_reply=True))
